
#include <fcntl.h>
#include <linux/fb.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <pthread.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include "hack.h"
#include "ui.h"

extern void UpdateGame(unsigned long ulButtons, unsigned long ulJoystick);
extern void RenderScreen(unsigned short *pus, unsigned long ulButtons);

static unsigned long g_hFrameHandles[2];
static unsigned short *g_pScreen[2];
static unsigned long g_ulScreenHardware[2];

unsigned long g_hMemory;
volatile unsigned long *g_pulRegs;
volatile unsigned short *g_pusRegs;

static void dzzInitScreen()
{
  struct fb_fix_screeninfo finfo;

  // open the system files
  g_hFrameHandles[0] = open("/dev/fb0", O_RDWR);
  g_hFrameHandles[1] = open("/dev/fb1", O_RDWR);
  g_hMemory = open("/dev/mem", O_RDWR);

  // set up machine register memory map
  g_pulRegs = (unsigned long  *) mmap(0, 0x10000, 
      PROT_READ|PROT_WRITE, MAP_SHARED, g_hMemory, 0xc0000000);
  g_pusRegs = (unsigned short *) g_pulRegs;

  // set up screen memory map.  the mmap gets us the virtual addresses
  // of the screens and the ioctl gets us the physical addresses
  ioctl(g_hFrameHandles[0], FBIOGET_FSCREENINFO, &finfo);
  g_ulScreenHardware[0] = finfo.smem_start;
  g_pScreen[0] = (unsigned short *) mmap(0, 153600, PROT_WRITE, MAP_SHARED,
      g_hFrameHandles[0], 0);

  ioctl(g_hFrameHandles[1], FBIOGET_FSCREENINFO, &finfo);
  g_ulScreenHardware[1] = finfo.smem_start;
  g_pScreen[1] = (unsigned short *) mmap(0, 153600, PROT_WRITE, MAP_SHARED,
      g_hFrameHandles[1], 0);

  g_pusRegs[0x290C>>1] = 640;
  g_pusRegs[0x28DA>>1] = 0x04AB;
}

void dzzCleanupHardware()
{
  munmap((unsigned long *) g_pulRegs, 0x10000);
  munmap(g_pScreen[0], 153600);
  munmap(g_pScreen[1], 153600);

  if(g_hFrameHandles[0])
    close(g_hFrameHandles[0]);
  if(g_hFrameHandles[1])
    close(g_hFrameHandles[1]);
  if(g_hMemory)
    close(g_hMemory);
}

int dzzIsVSync()
{
  if(g_pusRegs[0x1182>>1]&(1<<4))
    return 0;
  return 1;
}

void WaitForVSync()
{
  if(g_pusRegs[0x2804>>1] & 1)
  {
    while(g_pusRegs[0x1182>>1]&(1<<4))
      ;
  }
  else
  {
    while((g_pusRegs[0x1182>>1]&(1<<4)) == 0)
      ;
  }
}

void WaitForNotVSync()
{
  if(g_pusRegs[0x2804>>1] & 1)
  {
    while((g_pusRegs[0x1182>>1]&(1<<4)) == 0)
      ;
  }
  else
  {
    while(g_pusRegs[0x1182>>1]&(1<<4))
      ;
  }
}

static volatile unsigned short g_usCurFrame = 1;

void dzzFlipScreen()
{
  unsigned long ulAddr = (unsigned long) g_ulScreenHardware[g_usCurFrame];
  if(g_usCurFrame)
    g_usCurFrame = 0;
  else
    g_usCurFrame = 1;

  g_pusRegs[0x290E>>1] = g_pusRegs[0x2912>>1]=(unsigned short) (ulAddr&0xFFFF);
  g_pusRegs[0x2910>>1] = g_pusRegs[0x2914>>1]=(unsigned short)(ulAddr >> 16);
}

unsigned short *dzzGetScreen()
{
  return g_pScreen[g_usCurFrame];
}

#include "splash.c"

void QuickInitUI(int bDoSplash)
{
  dzzInitScreen();
  if(bDoSplash)
  {
    memcpy(dzzGetScreen(), g_usSplashData, 320*240*2);
    dzzFlipScreen();
    memcpy(dzzGetScreen(), g_usSplashData, 320*240*2);
    dzzFlipScreen();
  }
}

unsigned long dzzGetButtons()
{
  unsigned long value=(g_pusRegs[0x1198>>1] & 0x00FF);

  if(value==0xFD) value=0xFA;
  if(value==0xF7) value=0xEB;
  if(value==0xDF) value=0xAF;
  if(value==0x7F) value=0xBE;
 
  value =  ~((g_pusRegs[0x1184>>1] & 0xFF00) | value | (g_pusRegs[0x1186>>1] << 16));
  return value & 0x08C0FF00;
}

int dzzShiftDown()
{
  if(dzzGetButtons() & GP2X_L)
    return 1;
  return 0;
}

#define B_U  0x01
#define B_UL 0x02
#define B_L  0x04
#define B_DL 0x08
#define B_D  0x10
#define B_DR 0x20
#define B_R  0x40
#define B_UR 0x80

unsigned long dzzGetJoystick()
{
  unsigned long b = (~(g_pusRegs[0x1198>>1] & 0x00FF));

  if(b & B_U)
  {
    if(b & B_UL)
      return JOYDIR_UUL;
    if(b & B_UR)
      return JOYDIR_UUR;
    return JOYDIR_U;
  }
  if(b & B_L)
  {
    if(b & B_UL)
      return JOYDIR_LUL;
    if(b & B_DL)
      return JOYDIR_LDL;
    return JOYDIR_L;
  }
  if(b & B_D)
  {
    if(b & B_DL)
      return JOYDIR_DDL;
    if(b & B_DR)
      return JOYDIR_DDR;
    return JOYDIR_D;
  }
  if(b & B_R)
  {
    if(b & B_UR)
      return JOYDIR_RUR;
    if(b & B_DR)
      return JOYDIR_RDR;
    return JOYDIR_R;
  }
  if(b & B_UL)
    return JOYDIR_UL;
  if(b & B_UR)
    return JOYDIR_UR;
  if(b & B_DL)
    return JOYDIR_DL;
  if(b & B_DR)
    return JOYDIR_DR;
  return JOYDIR_NONE;
}

// convert the joystick direction to a "mode 2" 8-way

unsigned long NormalizeJoystick(unsigned long ulIn)
{
  switch(ulIn)
  {
    case JOYDIR_U:
    case JOYDIR_UUL:
    case JOYDIR_UUR:
      return JOYDIR_U;

    case JOYDIR_UL:
      return JOYDIR_UL;

    case JOYDIR_LUL:
    case JOYDIR_L:
    case JOYDIR_LDL:
      return JOYDIR_L;

    case JOYDIR_DL:
      return JOYDIR_DL;

    case JOYDIR_DDL:
    case JOYDIR_D:
    case JOYDIR_DDR:
      return JOYDIR_D;

    case JOYDIR_DR:
      return JOYDIR_DR;

    case JOYDIR_RDR:
    case JOYDIR_R:
    case JOYDIR_RUR:
      return JOYDIR_R;

    case JOYDIR_UR:
      return JOYDIR_UR;

    default:
      break;
  }
  return JOYDIR_NONE;
}

unsigned int g_ulFrameTimer = 0;

void dzzStartFrameTimer()
{
  g_ulFrameTimer = g_pulRegs[0x0A00>>2];
}

int dzzGetFrameTimer()
{
  unsigned long ul;

  ul = g_pulRegs[0x0A00>>2];
  if(ul > g_ulFrameTimer)
    return (int) ((ul - g_ulFrameTimer + 368) / 737);
  // wrapped around
  return (int) ((((ul/4)+(1<<30)) - (g_ulFrameTimer/4)) / 184);
}

unsigned int g_ulJoystickTimer = 0;

void dzzStartJoystickTimer()
{
  g_ulJoystickTimer = g_pulRegs[0x0A00>>2];
}

int dzzGetJoystickTimer()
{
  unsigned long ul;

  ul = g_pulRegs[0x0A00>>2];
  if(ul > g_ulJoystickTimer)
    return (int) ((ul - g_ulJoystickTimer + 368) / 737);
  // wrapped around
  return (int) ((((ul/4)+(1<<30)) - (g_ulJoystickTimer/4)) / 184);
}

unsigned int g_ulStickClickTimer = 0;

void dzzStartStickClickTimer()
{
  g_ulStickClickTimer = g_pulRegs[0x0A00>>2];
}

int dzzGetStickClickTimer()
{
  unsigned long ul;

  ul = g_pulRegs[0x0A00>>2];
  if(ul > g_ulStickClickTimer)
    return (int) ((ul - g_ulStickClickTimer + 368) / 737);
  // wrapped around
  return (int) ((((ul/4)+(1<<30)) - (g_ulStickClickTimer/4)) / 184);
}

void dzzMicroSleep(int nMicros)
{
  usleep((unsigned long) nMicros);
}

static volatile int g_bUIThreadRunning = 0;
static volatile int g_bExitUIThread = 0;

static void *UIThread(void *p)
{
  unsigned long ulButtons;
  unsigned long ulJoystick;
  // int bIsVSync;

  g_bUIThreadRunning = 1;
  // set up stuff
  // dzzInitScreen(); // this is done int QuickInitUI now
  
  while(!g_bExitUIThread)
  {
    // do any game updating.  About all we'd do here is
    // autoscroll the map
    dzzStartFrameTimer();
    ulButtons = dzzGetButtons();
    ulJoystick = dzzGetJoystick();
    ulJoystick = NormalizeJoystick(ulJoystick);
    UpdateGame(ulButtons, ulJoystick);

    // draw the screen buffer
    // using the memscreen and current state, etc
    RenderScreen(dzzGetScreen(), ulButtons);
    // wait for vertical sync
    while(dzzGetFrameTimer() < 200)
      dzzMicroSleep(5000);
    /*
    bIsVSync = dzzIsVSync();
    while(1)
    {
      if(dzzIsVSync() != bIsVSync)
        break;
    }
    */
    WaitForNotVSync();
    WaitForVSync();
    // flip the screen
    dzzFlipScreen();
  }
  dzzCleanupHardware();
  g_bUIThreadRunning = 0;
  return 0;
}

void StartUIThread()
{
  pthread_t thr;
  pthread_create(&thr, NULL, UIThread, NULL);
}

void ShutdownUIThread()
{
  g_bExitUIThread = 1;
  while(g_bUIThreadRunning)
  {
    usleep(10000);
  }
}
