#include <stdio.h>
#include <stdlib.h>
#include <png.h>

#define XRGB(r,g,b) (((r) << 11) | ((g) << 5) | b)

unsigned short g_usTiles32[1280*960];
unsigned short g_usTiles8[160*240];

extern unsigned short g_usTransparentColor;
extern unsigned short g_usBackgroundColor;
extern unsigned short g_usCursorColor;
extern int g_bTransparent;

extern int g_nMediumTileWidth;
extern int g_nMediumTileHeight;
extern int g_nLargeTileWidth;
extern int g_nLargeTileHeight;

extern unsigned short g_usDefaultTiles16[640*480];
unsigned short g_usTiles16[640*480];

int LoadTiles(char *pszFile, unsigned short *pus, int nMaxPixels)
{
  FILE *fp;
  png_structp png_ptr;
  png_infop info_ptr;
  unsigned char *image_data;
  char sig[8];
  int bit_depth;
  int color_type;
  unsigned long width;
  unsigned long height, i;
  unsigned int rowbytes;
  png_bytepp row_pointers = NULL;
  unsigned short *pusLoop;

  fp = fopen(pszFile, "r");
  if(!fp)
  {
    return 0;
  }
  fread(sig, 1, 8, fp);
  if(!png_check_sig((unsigned char *) sig, 8))
  {
    fclose(fp);
    return 0;
  }
  png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  if(!png_ptr)
  {
    fclose(fp);
    return 0;
  }
  info_ptr = png_create_info_struct(png_ptr);
  if(!info_ptr)
  {
    png_destroy_read_struct(&png_ptr, NULL, NULL);
    fclose(fp);
    return 0;
  }
  if(setjmp(png_jmpbuf(png_ptr)))
  {
    png_destroy_read_struct(&png_ptr, NULL, NULL);
    fclose(fp);
    return 0;
  }
  png_init_io(png_ptr, fp);
  png_set_sig_bytes(png_ptr, 8);
  png_read_info(png_ptr, info_ptr);
  png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type, NULL, NULL, NULL);

  if (color_type & PNG_COLOR_MASK_ALPHA)
      png_set_strip_alpha(png_ptr);
  if (bit_depth > 8)
      png_set_strip_16(png_ptr);
  if (color_type == PNG_COLOR_TYPE_GRAY || color_type == PNG_COLOR_TYPE_GRAY_ALPHA)
     png_set_gray_to_rgb(png_ptr);
  if (color_type == PNG_COLOR_TYPE_PALETTE)
     png_set_palette_to_rgb(png_ptr);
   
  png_read_update_info(png_ptr, info_ptr);
  rowbytes = png_get_rowbytes(png_ptr, info_ptr);

  /* Allocate the image_data buffer. */
  if ((image_data = (unsigned char *) malloc(rowbytes * height))==NULL)
  {
    png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
    fclose(fp);
    return 0;
  }

   if ((row_pointers = (png_bytepp)malloc(height*sizeof(png_bytep))) == NULL)
   {
     png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
     free(image_data);
     fclose(fp);
     return 0;
   }

   for (i = 0;  i < height;  ++i)
     row_pointers[height - 1 - i] = image_data + i*rowbytes;
   png_read_image(png_ptr, row_pointers);

   pusLoop = pus;
  {
    int nPixels, nRed, nGreen, nBlue, x, y;
    nPixels = 0;
    for(y = 0; y < height; y++)
    {
     if(nPixels >= nMaxPixels)
       break;
     for(x = 0; x < width; x++)
     {
       if(nPixels >= nMaxPixels)
        break;
       nRed = row_pointers[y][x*3] >> 3;
       nGreen = row_pointers[y][x*3+1] >> 2;
       nBlue = row_pointers[y][x*3+2] >> 3;
       *pusLoop = (unsigned short) ((nRed<<11)|(nGreen<<5)|(nBlue));
       pusLoop++;
       nPixels++;
     }
    }
  }

  // get image properties
  g_usTransparentColor = pus[width*(height-3)+width-1];
  g_usBackgroundColor = pus[width*(height-1)+width-1];
  g_usCursorColor = pus[width*(height-4)+width-1];
  if(pus[width*(height-2)+width-1] == 0)
    g_bTransparent = 1;
  else
    g_bTransparent = 0;
  g_nLargeTileWidth = width / 40;
  g_nLargeTileHeight = height / 27;
  g_nMediumTileWidth = g_nLargeTileWidth / 2;
  g_nMediumTileHeight = g_nLargeTileHeight / 2;

  free(row_pointers);
  free(image_data);
  png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
  fclose(fp);
  return 1;
}

extern char g_szTileFile[200];

int InitTiles32()
{
  int nX, nY, x, y, nRed, nGreen, nBlue;
  unsigned short us;
  int nReturn = 1;

  if(g_szTileFile[0])
  {
    if(LoadTiles(g_szTileFile, g_usTiles32, 1280*960))
    {
      for(nY = 0; nY < g_nMediumTileHeight*27; nY++)
      {
        for(nX = 0; nX < g_nMediumTileWidth*40; nX++)
        {
          nRed = 0;
          nGreen = 0;
          nBlue = 0;
          for(y = 0; y < 2; y++)
          {
            for(x = 0; x < 2; x++)
            {
              us = g_usTiles32[(nY*2+y)*(g_nLargeTileWidth*40)+(nX*2+x)];
              nRed = nRed + (us >> 11);
              nGreen = nGreen + ((us>>5) & 0x3F);
              nBlue = nBlue + (us & 0x1F);
            }
          }
          nRed = (nRed + 2) / 4;
          nGreen = (nGreen + 2) / 4;
          nBlue = (nBlue + 2) / 4;
          us = (unsigned short) ((nRed << 11) | (nGreen << 5) | nBlue);
          g_usTiles16[nY*(g_nMediumTileWidth*40)+nX] = us;
        }
      }
      return 1;
    }
    nReturn = 0;
  }

  // if we get here, we want to use the defaults
  g_usTransparentColor = 0;
  g_usBackgroundColor = 0;
  g_usCursorColor = XRGB(31,63,10);
  g_bTransparent = 0;
  g_nLargeTileWidth = 32;
  g_nLargeTileHeight = 32;
  g_nMediumTileWidth = 16;
  g_nMediumTileHeight = 16;
  for(nY = 0; nY < g_nMediumTileHeight*27; nY++)
  {
    for(nX = 0; nX < g_nMediumTileWidth*40; nX++)
    {
      g_usTiles16[nY*g_nMediumTileWidth*40+nX] = g_usDefaultTiles16[nY*(g_nMediumTileWidth*40)+nX];
    }
  }
  g_szTileFile[0] = 0;
  for(nY = 0; nY < g_nMediumTileHeight*27; nY++)
  {
    for(nX = 0; nX < g_nMediumTileWidth*40; nX++)
    {
      us = g_usTiles16[nY*g_nMediumTileWidth*40+nX];
      g_usTiles32[nY*2*g_nLargeTileWidth*40+nX*2] = us;
      g_usTiles32[nY*2*g_nLargeTileWidth*40+nX*2+1] = us;
      g_usTiles32[(nY*2+1)*g_nLargeTileWidth*40+nX*2] = us;
      g_usTiles32[(nY*2+1)*g_nLargeTileWidth*40+nX*2+1] = us;
    }
  }
  return nReturn;
}


// reduces an arbitrary (larger) tile to 4x8 for the map

void ReduceTo4x8(int nSrcWidth, int nSrcHeight, unsigned short *pusSrc, int nSrcStride,
		unsigned short *pusDest)
{
	int nRedVals[4*8];
	int nGreenVals[4*8];
	int nBlueVals[4*8];
	int nX, nY;
	int xFraction, xFractionX, xFractionY, nRed, nGreen, nBlue;
  int nDestX, nDestY;
	unsigned short us;

	memset(nRedVals, 0, 4 * 8 * 4);
	memset(nGreenVals, 0, 4 * 8 * 4);
	memset(nBlueVals, 0, 4 * 8 * 4);
	for(nY = 0; nY < nSrcHeight; nY++)
	{
		for(nX = 0; nX < nSrcWidth; nX++)
		{
			nDestX = nX * 4 / nSrcWidth;
			nDestY = nY * 8 / nSrcHeight;
			// what fraction of the current source pixel is inside the
			// destination pixel?
			xFractionX = ((((nDestX+1) * 4096) - (nX * 4096 * 4 / nSrcWidth)) *
				nSrcWidth) / 4;
			if(xFractionX > 4096)
				xFractionX = 4096;
			xFractionY = ((((nDestY+1) * 4096) - (nY * 4096 * 8 / nSrcHeight)) *
				nSrcHeight) / 8;
			if(xFractionY > 4096)
				xFractionY = 4096;
			// now update the four possible pixels
			us = pusSrc[nY * nSrcStride + nX];
			nRed = us >> 11;
			nGreen = (us >> 5) & 0x3F;
			nBlue = us & 0x1F;
      /*
			nRedVals[nDestY*4+nDestX] += (nRed<<16);
			nGreenVals[nDestY*4+nDestX] += (nGreen<<16);
			nBlueVals[nDestY*4+nDestX] += (nBlue<<16);
      */
			nRedVals[nDestY*4+nDestX] += (((xFractionX * xFractionY) >> 8) * nRed);
			nGreenVals[nDestY*4+nDestX] += (((xFractionX * xFractionY) >> 8) * nGreen);
			nBlueVals[nDestY*4+nDestX] += (((xFractionX * xFractionY) >> 8) * nBlue);
			if((xFractionX < 4096) && (nDestX+1 < 4))
			{
				nRedVals[nDestY*4+nDestX+1] += ((((4096 - xFractionX) * xFractionY) >> 8) * nRed);
				nGreenVals[nDestY*4+nDestX+1] += ((((4096 - xFractionX) * xFractionY) >> 8) * nGreen);
				nBlueVals[nDestY*4+nDestX+1] += ((((4096 - xFractionX) * xFractionY) >> 8) * nBlue);
			}
			if((xFractionY < 4096) && (nDestY+1 < 8))
			{
				nRedVals[(nDestY+1)*4+nDestX] += (((xFractionX * (4096 - xFractionY)) >> 8) * nRed);
				nGreenVals[(nDestY+1)*4+nDestX] += (((xFractionX * (4096 - xFractionY)) >> 8) * nGreen);
				nBlueVals[(nDestY+1)*4+nDestX] += (((xFractionX * (4096 - xFractionY)) >> 8) * nBlue);
				if((xFractionX < 4096) && (nDestX+1 < 4))
				{
					// nRedVals etc are in .12 format
					nRedVals[(nDestY+1)*4+nDestX+1] += ((((4096 - xFractionX) * (4096 - xFractionY)) >> 8) * nRed);
					nGreenVals[(nDestY+1)*4+nDestX+1] += ((((4096 - xFractionX) * (4096 - xFractionY)) >> 8) * nGreen);
					nBlueVals[(nDestY+1)*4+nDestX+1] += ((((4096 - xFractionX) * (4096 - xFractionY)) >> 8) * nBlue);
				}
			}
		}
	}
	// now translate the results into resulting colors
	// xFraction is the reciprocal of the number of source pixels per dest pixel, in .8 format.
	// Multiply by this to adjust the color value
	// The highest ratio we're concerned with is 48x48 -> 4x8 which is 72. (7 bits).  Green
	// takes 6 bits, making 13, plus 1 for the sign bit makes 14, leaving 18 bits of overhead
	// for the fixed point, so convert the values to .10 before the multiplication.
	xFraction = (4 * 8 * 256) / (nSrcWidth * nSrcHeight);
	for(nY = 0; nY < 8; nY++)
	{
		for(nX = 0; nX < 4; nX++)
		{
			nRed = ((nRedVals[nY*4+nX]>>2)*xFraction)>>22;
			nGreen = ((nGreenVals[nY*4+nX]>>2)*xFraction)>>22;
			nBlue = ((nBlueVals[nY*4+nX]>>2)*xFraction)>>22;
			us = (unsigned short) ((nRed << 11) + (nGreen << 5) + (nBlue));
			pusDest[nY*4*40+nX] = us;
			// pusDest[nY*4*40+nX] = 0x1F;
		}
	}
}

void InitTiles8()
{
  int nX, nY;
  for(nY = 0; nY < 27; nY++)
  {
    for(nX = 0; nX < 40; nX++)
    {
      ReduceTo4x8(g_nMediumTileWidth, g_nMediumTileHeight,
                  &g_usTiles16[nY*g_nMediumTileHeight*g_nMediumTileWidth*40+nX*g_nMediumTileWidth], g_nMediumTileWidth*40, &g_usTiles8[nY*8*160+nX*4]);
    }
  }
}

